/**
 * Created by Ludo Pulles on 2017-12-26.
 */

/*******************************************************************************
 * Zuniq is the object which manages the gomoku visualisation.
 * This part makes sure that the board transitions can be made.
 * When the board is changed, it sends the board to all its controllers.
 ******************************************************************************/

function Zuniq(playerNames, moves) {
	this.playerNames = playerNames;
	charA = 'A'.charCodeAt(0);
	char1 = '1'.charCodeAt(0);
	this.size = 5;
	var self = this;

	this.moves = [ "start" ];
	for (var i in moves) {
		var row = moves[i].charCodeAt(0) - charA, col = moves[i].charCodeAt(1) - char1;
		this.moves.push({ x: row, y: col, hor: moves[i].charAt(2) == 'h' });
	}

	// determine final areas...
	this.finalAreas = [];

	var dx = [ -1, 0, 1, 0 ], dy = [ 0, 1, 0, -1 ];
	var color = [];
	for (var x = 0; x < this.size; x++) {
		var tmp = [];
		for (var y = 0; y < this.size; y++)
			tmp.push(0);
		color.push(tmp);
	}

	var max = Math.max;

	var find_hor = function(x, y) {
		var opt1 = moves.indexOf(String.fromCharCode(charA + x, char1 + y) + 'h');
		var opt2 = moves.indexOf(String.fromCharCode(charA + x, char1 + y) + 'h!');
		return Math.max(opt1, opt2);
	};

	var find_ver = function(x, y) {
		var opt1 = moves.indexOf(String.fromCharCode(charA + x, char1 + y) + 'v');
		var opt2 = moves.indexOf(String.fromCharCode(charA + x, char1 + y) + 'v!');
		return Math.max(opt1, opt2);
	};

	this.areas = {};
	for (var x = 0; x < this.size; x++) {
		for (var y = 0; y < this.size; y++) {
			if (color[x][y] != 0) continue;
			
			var cells = [], last_idx = 0;
			var q = [], qst = 0, border = false;
			q.push([x,y]);
			while (qst < q.length) {
				var xx = q[qst][0], yy = q[qst][1]; qst++;

				if (xx < 0 || yy < 0 || xx >= this.size || yy >= this.size)
					{ border = true; continue; }
				if (color[xx][yy] != 0) continue;
				color[xx][yy] = -1;
				cells.push([xx,yy]);

				var idx;
				
				idx = find_hor(xx, yy);
				if (idx < 0) q.push([xx-1,yy]);
				else last_idx = max(last_idx, idx);
				
				idx = find_hor(xx+1, yy);
				if (idx < 0) q.push([xx+1,yy]);
				else last_idx = max(last_idx, idx);

				idx = find_ver(xx, yy);
				if (idx < 0) q.push([xx,yy-1]);
				else last_idx = max(last_idx, idx);

				idx = find_ver(xx, yy+1);
				if (idx < 0) q.push([xx,yy+1]);
				else last_idx = max(last_idx, idx);
			}

			if (border) continue;
/*
			var sz = cells.length;
			for (var i in cells) {
				color[ cells[i][0] ][ cells[i][1] ] = sz;
			}
*/
			this.areas[last_idx] = cells;
		}
	}

	this.isRunning = false;
	this.board = document.getElementById('game-container');
	this.moveTable = document.getElementById('move-table').children[0];

	var showLegends = true;

	// create the board
	var table = document.createElement('table');
	var thead = document.createElement('thead');
	var thr = document.createElement('tr');

	this.rows = [];
	this.cols = [];

	if (showLegends) {
		// thr.appendChild(document.createElement('th'));

		for (var col = 0; col <= this.size; col++) {
			var th = document.createElement('th');

			if (col == 0) {
				var div = document.createElement('div');
				div.innerText = "A";
				div.className = "row-legend";
				th.appendChild(div);
				this.rows.push(th);
			}

			var div = document.createElement('div');
			div.innerText = String.fromCharCode(char1 + col);
			div.className = "col-legend";
			th.appendChild(div);
			this.cols.push(th);

			thr.appendChild(th);
		}
		thead.appendChild(thr);
		table.appendChild(thead);
	}

	this.cells = [];

	var tbody = document.createElement('tbody');
	for (var row = 0; row < this.size; row++) {
		var tr = document.createElement('tr');

		var cellRow = [];

		if (showLegends) {
			var rowName = String.fromCharCode(charA + row + 1);
			var rowElem = document.createElement('td');
			var div = document.createElement('div');
			div.innerText = rowName;
			div.className = "row-legend";
			rowElem.appendChild(div);
			rowElem.className = "frow";
			tr.appendChild(rowElem);
			this.rows.push(rowElem);
		}
		for (var col = 0; col < this.size; col++) {
			var cell = document.createElement('td');
			cell.className = "game-cell";

			tr.appendChild(cell);
			cellRow.push(cell);
		}

		this.cells.push(cellRow);
		tbody.appendChild(tr);
	}

	var X = document.createElement('tr');
	tbody.appendChild(X);
	var btnwrapper = document.createElement('td');
	btnwrapper.setAttribute('colspan', this.size + 1);
	btnwrapper.setAttribute('style', 'text-align:center;');
	btnwrapper.appendChild(document.getElementsByClassName('button-bar')[0]);
	X.appendChild(btnwrapper);
	table.appendChild(tbody);

	this.board.appendChild(table);

	this.curMove = 0;

	this.registerEvents();
	this.buildMoveTable(moves);
	this.updateSelectedRow();
}

// returns whether the move has been processed
Zuniq.prototype.setMove = function(move) {
	if (move < 0 || move >= this.moves.length) return false;
	if (move == this.curMove) return true;
	var self = this;

	changeClass = function(x, y, dir) {

		var toggle = function(node, cl) {
			var classes = node.className.split(" ");
			if (classes.indexOf(cl) != -1)
				classes.splice(classes.indexOf(cl), 1);
			else
				classes.push(cl);
			node.className = classes.join(" ");
		};

		dir = 'bar-' + dir;

		if (y == -1)
			toggle(self.rows[x+1], dir);
		else if (x == -1)
			toggle(self.cols[y+1], dir);
		else
			toggle(self.cells[x][y], dir);
	};

	processMove = function(m, mv_nr, add) {
		// x: row, y: col
		if (m.hor) {
			// if (m.x > 0)
			changeClass(m.x-1, m.y, 'bottom', add);
			if (m.x < self.size)
				changeClass(m.x, m.y, 'top', add);
		} else {
			// if (m.y > 0)
			changeClass(m.x, m.y-1, 'right', add);
			if (m.y < self.size)
				changeClass(m.x, m.y, 'left', add);
		}

		var v = self.areas[mv_nr];
		if (v !== undefined) {
			for (var i in v) {
				var table_cell = self.cells[v[i][0]][v[i][1]];
				table_cell.innerHTML = add ? v.length : '';
			}
		}
	};

	while (this.curMove > move) {
		// change a move back.
		// undo this.curMove
		var cur = this.moves[this.curMove--];
		processMove(cur, this.curMove, false);
	}
	while (this.curMove < move) {
		// process this.curMove
		var cur = this.moves[++this.curMove];
		processMove(cur, this.curMove-1, true);
	}

	this.updateSelectedRow();
	return true;
};

// -----------------------------------------------------------------------------
// Event handlers:
// -----------------------------------------------------------------------------

Zuniq.prototype.setRunning = function(running) {
	if (this.isRunning == running) return;

	this.isRunning = running;
	this.playPause.innerText = running ? 'Pause' : 'Play';

	var self = this, run = function() {
		if (!self.isRunning) return;

		if (self.setMove(self.curMove + 1)) {
			setTimeout(run, 500);
		} else {
			// last move has been reached.
			self.setRunning(false);
		}
	}
	if (this.isRunning) run();
}

Zuniq.prototype.prevMove = function() { this.setRunning(false); this.setMove(this.curMove - 1); };
Zuniq.prototype.nextMove = function() { this.setRunning(false); this.setMove(this.curMove + 1); };

Zuniq.prototype.registerEvents = function() {
	this.playPause = document.getElementById('play-pause');

	var self = this;
	document.getElementById('first').onclick = function() { self.setRunning(false); self.setMove(0); };
	document.getElementById('last').onclick = function() { self.setRunning(false); self.setMove(self.moves.length - 1); };
	document.getElementById('prev').onclick = function() { self.prevMove(); };
	document.getElementById('next').onclick = function() { self.nextMove(); };
	this.playPause.onclick = function() { self.setRunning(!self.isRunning); };
};


Zuniq.prototype.buildMoveTable = function(moves) {
	var table = document.createElement('table'), self = this;
	var theader = document.createElement('tr');
	theader.appendChild(document.createElement('th'));
	for (var i = 0; i < 2; i++) {
		var th = document.createElement('th');

		// cut-off on more than 20 characters in the name
		var playerName = i == 0 ? "white" : "black";
		if (this.playerNames) {
			var name = this.playerNames[i];
			if (name.length > 20) name = name.slice(0, 17) + "...";

			playerName = name + " as " + playerName;
		}

		th.innerText = playerName;
		theader.appendChild(th);
	}

	var thead = document.createElement('thead');
	thead.appendChild(theader);
	table.appendChild(thead);

	var tbody = document.createElement('tbody');

	for (var i = 0; i <= moves.length; i++) {
		var tr = document.createElement('tr');
		var moveNr = document.createElement('th');
		moveNr.innerText = i + '.';
		if (i % 2) tr.setAttribute('class', 'zebra');
		tr.appendChild(moveNr);

		var curp = !i ? 0 : ~i & 1;
		for (var j = 0; j < 2; j++) {
			var td = document.createElement('td');
			if (j == curp) td.innerText = !i ? "Start" : moves[i - 1];
			tr.appendChild(td);
		}

		tr.onclick = (function(move) {
			return function() { self.setMove(move); }
		})(i);

		tbody.appendChild(tr);
	}

	this.moveTableBody = tbody;
	this.moveTable.onkeydown = function(event) {
		if (event.keyCode == 38 || event.keyCode == 37) {
			self.prevMove();
		} else if (event.keyCode == 40 || event.keyCode == 39) {
			self.nextMove();
		} else {
			return;
		}
		event.preventDefault();
		event.stopImmediatePropagation();
	};

	table.appendChild(tbody);
	this.moveTable.appendChild(table);
};

Zuniq.prototype.updateSelectedRow = function() {
	var lastCurMove = document.getElementById('cur-move');
	var tbl = this.moveTable;

	// reset the id.
	if (lastCurMove) lastCurMove.id = '';

	if (this.curMove >= 0) {
		var curRow = this.moveTableBody.children[this.curMove];
		curRow.id = 'cur-move';

		var topCoord = curRow.offsetTop;
		var bottomCoord = curRow.offsetTop + curRow.offsetHeight - tbl.offsetHeight;
		if (tbl.scrollTop < bottomCoord) {
			tbl.scrollTop = bottomCoord;
		} else if (tbl.scrollTop > topCoord) {
			tbl.scrollTop = topCoord;
		}
	} else {
		tbl.scrollTop = 0;
	}
};
